﻿using System;
using System.Collections.Generic;
using System.Dynamic;
using System.Linq;
using Microsoft.WindowsAzure.Storage;
using Microsoft.WindowsAzure.Storage.Table;
using Newtonsoft.Json;

namespace SalesAnalytics
{
    public class TableEntity : DynamicObject, ITableEntity
    {
        #region ITableEntity properties

        public string ETag { get; set; }

        public string PartitionKey { get; set; }

        public string RowKey { get; set; }

        public DateTimeOffset Timestamp { get; set; }

        #endregion

        private IDictionary<string, EntityProperty> _properties = new Dictionary<string, EntityProperty>();

        #region override DynamicObject's methods

        public override bool TryGetMember(GetMemberBinder binder, out object result)
        {
            if (!_properties.ContainsKey(binder.Name))
                _properties.Add(binder.Name, ConvertToEntityProperty(binder.Name, null));
            result = _properties[binder.Name];
            return true;
        }

        public override bool TrySetMember(SetMemberBinder binder, object value)
        {
            var property = ConvertToEntityProperty(binder.Name, value);

            if (_properties.ContainsKey(binder.Name))
                _properties[binder.Name] = property;
            else
                _properties.Add(binder.Name, property);

            return true;
        }

        public bool TrySetMember(string binder, object value)
        {
            var property = ConvertToEntityProperty(binder, value);

            if (_properties.ContainsKey(binder))
                _properties[binder] = property;
            else
                _properties.Add(binder, property);

            return true;
        }

        #endregion

        #region ITableEntity implementation

        public void ReadEntity(IDictionary<string, EntityProperty> properties, OperationContext operationContext)
        {
            _properties = properties;
        }

        public IDictionary<string, EntityProperty> WriteEntity(OperationContext operationContext)
        {
            return _properties;
        }

        #endregion

        private static EntityProperty ConvertToEntityProperty(string key, object value)
        {
            if (value == null) return new EntityProperty((string) null);
            if (value.GetType() == typeof (byte[]))
                return new EntityProperty((byte[]) value);
            if (value is bool)
                return new EntityProperty((bool) value);
            if (value is DateTimeOffset)
                return new EntityProperty((DateTimeOffset) value);
            if (value is DateTime)
                return new EntityProperty((DateTime) value);
            if (value is double)
                return new EntityProperty((double) value);
            if (value is Guid)
                return new EntityProperty((Guid) value);
            if (value is int)
                return new EntityProperty((int) value);
            if (value is long)
                return new EntityProperty((long) value);
            var s = value as string;
            if (s != null)
                return new EntityProperty(s);
            throw new Exception(string.Format("This value type {0} is not supported for {1}", key, value));
        }

        public static TableEntity CreateTableEntity<T>(
            T entity,
            Func<T, string> getPartitionKey,
            Func<T, string> getRowKey)
            where T : class
        {
            var tableEntity = new TableEntity
            {
                PartitionKey = getPartitionKey(entity),
                RowKey = getRowKey(entity),
                Timestamp = DateTime.Now
            };

            //now set all the entity properties
            var type = typeof (T);
            foreach (var p in type.GetProperties())
            {
                var pt = p.PropertyType;

                if (
                    pt.IsGenericType && typeof (ICollection<>).IsAssignableFrom(pt.GetGenericTypeDefinition())
                    ||
                    pt.GetInterfaces()
                        .Any(x => x.IsGenericType && x.GetGenericTypeDefinition() == typeof (ICollection<>))
                    ||
                    pt.IsClass
                    )
                    tableEntity.TrySetMember(p.Name, JsonConvert.SerializeObject(p.GetValue(entity, null)));
                else
                    tableEntity.TrySetMember(p.Name, p.GetValue(entity, null) ?? "");
            }

            return tableEntity;

        }

    }

}
